#!/usr/bin/env python
"""Visualization of the statistics."""
import numpy as np
import matplotlib.pyplot as plt

def dist_signal(
    img,
    threshold: float=None,
    label_threshold: float=None,
    params_axvline: dict={'color':'r','linestyle':'dashed'},
    ax: plt.Axes=None,
    **kws,
    ) -> plt.Axes:
    """Plot the distribution of intensity.

    Args:
        img (_type_): inpput image
        threshold (float, optional): threshold applied. Defaults to None.
        label_threshold (float, optional): label of the threshold. Defaults to None.
        params_axvline (_type_, optional): parameters provided to the vertical line plot. Defaults to {'color':'r','linestyle':'dashed'}.
        ax (plt.Axes, optional): subplot object. Defaults to None.

    Keyword Args:
        parameters provided to the `hist` function. 
    
    Returns:
        plt.Axes
    """
    ax=plt.subplot() if ax is None else ax 
    a=np.ravel(img)
    a = a[~np.isnan(a)]
    a=np.extract(a<np.quantile(a,0.9),a)
    _=ax.hist(a,linewidth=0,**kws,)
    # ax.set_xlim(np.min(a),np.quantile(a,0.9))
    if not threshold is None:
        ax.axvline(threshold,label=label_threshold,**params_axvline)
        ax.legend()
    ax.set_xlabel('signal')
    ax.set_ylabel('density')
    return ax

def plot_summary_stats(
    input_paths: list,
    ax: plt.Axes=None,
    ) -> plt.Axes:
    """Plot summary stats for a set of images e.g. time-lapse images.

    Args:
        input_paths (list): list of paths of the images.
        ax (plt.Axes, optional): subplot object. Defaults to None.

    Returns:
        plt.Axes
    """
    import pandas as pd
    from htsimaging.lib.io import read_image
    
    if ax is None:
        fig,ax=plt.subplots(figsize=[3,3])
    frames = [read_image(p) for p in input_paths]
    dframes=pd.DataFrame({i:np.ravel(np.array(frame)) for i,frame in enumerate(frames)})
    df=dframes.describe().T
    ax=df.loc[:,['mean','min','50%']].plot(ax=ax)
    return ax



#!/usr/bin/env python
"""Processing of the paths of input images to create configurations and metadata."""

import logging
from glob import glob
from os import makedirs
from os.path import isdir,abspath,exists,basename,dirname
import numpy as np
import matplotlib.pyplot as plt # used for QC plots
from roux.lib.io import read_dict, to_dict

def make_project_cfg(
    prjd: str,
    output_dir_path: str,
    bright_fn_marker: str=None,
    segmented_fn_marker:str=None,
    magnification: int=None,
    image_ext: str='tif',
    cores: int=1,
    test: bool=False,
    force: bool=False,
    )-> dict:
    """Make the confguration for the analysis run.

    Args:
        prjd (str): path to the directory with the images.
        output_dir_path (str): output directory path.
        bright_fn_marker (_type_): marker in the path of the bright field images.
        segmented_fn_marker (_type_): marker in the path of the segmented images.
        cores (int, optional): number of cores. Defaults to 1.
        test (bool, optional): test-mode. Defaults to False.
        force (bool, optional): over-write theoutputs. Defaults to False.

    Returns:
        dict: metadata
        
    Notes:
        Infer the magnification from the filenames:
        if 'T1C1' in cfg['bright_fn_marker']:
            cfg['magnification']=150
        elif cfg['bright_fn_marker']=='_t':
            cfg['magnification']=100
    """
    prjd=abspath(prjd)
    cfgp=f"{prjd}/cfg.yml"
    if not exists(cfgp) or force:    
        logging.info('making cfg')
        cfg={'prjd':prjd}
        cfg['outd']=output_dir_path
        cfg['cfgp']=cfgp
        cfg['cores']=cores
        cfg['bright_fn_marker']=bright_fn_marker
        cfg['segmented_fn_marker']=segmented_fn_marker
        assert not magnification is None
        cfg['magnification']=magnification
        # else:
        #     logging.error("unknown bright_fn_marker {cfg['bright_fn_marker']}")
        cfg['trials']={basename(d):{'datad':d} for d in glob(f"{cfg['prjd']}/*") if (isdir(d) and basename(d).replace('/','')!='segmentation_cell' and not basename(d).startswith('_'))}
        trials_bad=[]
        for k in cfg['trials']:
            if not cfg['bright_fn_marker'] is None:
                cfg['trials'][k]['bright']=[abspath(p) for p in glob(f"{cfg['trials'][k]['datad']}/*{image_ext}") if cfg['bright_fn_marker'] in p and not p.endswith(f'.segmented.{image_ext}')]
            else:
                cfg['trials'][k]['bright']=[]
            if not cfg['segmented_fn_marker'] is None:            
                cfg['trials'][k]['segmented']=[abspath(p) for p in glob(f"{cfg['trials'][k]['datad']}/*{image_ext}") if cfg['segmented_fn_marker'] in p]
            else:
                cfg['trials'][k]['segmented']=[]
            cfg['trials'][k]['gfp']=[abspath(p) for p in glob(f"{cfg['trials'][k]['datad']}/*{image_ext}")]
            cfg['trials'][k]['gfp']=list(set(cfg['trials'][k]['gfp'])-set(cfg['trials'][k]['bright']+cfg['trials'][k]['segmented']))
            cfg['trials'][k]['outd']=f"{cfg['outd']}/{k}/"
            cfg['trials'][k]['plotd']=f"{cfg['trials'][k]['outd']}/plot"
            makedirs(cfg['trials'][k]['plotd'],exist_ok=True)
            if len(cfg['trials'][k]['gfp'])==0:
                logging.warning(f"no gfp images for trial:{k}")
                trials_bad.append(k)
                continue
            if len(cfg['trials'][k]['bright'])+len(cfg['trials'][k]['segmented'])==0:
                logging.warning(f"no bright-field/segmented images for trial:{k}")
                trials_bad.append(k)
                continue
        for trial in trials_bad:
            logging.warning(f"removing trial:{trial}")
            del cfg['trials'][trial]
        to_dict(cfg,cfgp)   
        # QC
        ## 1 CHECK BLEACHING
        for k in cfg['trials']:
            from htsimaging.viz.stat import plot_summary_stats
            ax=plot_summary_stats(
                input_paths=sorted(cfg['trials'][k]['gfp'])
            )
            ax.set(xlabel='time points',ylabel='intensity',title="QC: bleaching test")
            plotp=f"{cfg['trials'][k]['plotd']}/plot_check_bleaching.png"
            makedirs(dirname(plotp),exist_ok=True)
            plt.savefig(plotp)
    else:
        cfg=read_dict(cfgp)
    return cfg

def make_cell_cfg(
    cfg: dict,
    frames: list,
    cells: list,
    trial: str,
    celli: int,
    cellbox: list,
    params_get_signal_summary_by_roi: dict={
        'xy_center':None,
        'width':20,
        'fun_summary_frame':'min',
        'fun_summary_frames':'median'
        },
    filterby_centroid: bool=False,
    scale_signal_cytoplasm: float=1.5,
    test: bool=False,
    force: bool=False,
    ) -> dict:
    """Make the configuration for an individual cell.

    Args:
        cfg (dict): metadata.
        frames (list): list of frames.
        cells (list): list of cells.
        trial (str): trial name.
        celli (int): index of the cell.
        cellbox (list): bounding box of the cell
        params_get_signal_summary_by_roi (dict, optional): parameters for the aggregation of the values at the ROI. Defaults to {'xy_center':None,'width':20, 'fun_summary_frame':'min', 'fun_summary_frames':'median' }.
        test (bool, optional): test-mode. Defaults to False.
        force (bool, optional): over-write the output. Defaults to False.

    Returns:
        dict: metadata
    """
    outp=f"{cfg['trials'][trial]['outd']}/cells/cell{celli:08d}/"
    cfgp=f"{outp}/cfg.yml"
    
    if not exists(cfgp) or force:
        cellcfg={}
        cellcfg['outp']=outp
        cellcfg['plotp']=f"{outp}/plot"
        cellcfg['cfgp']=cfgp
        cellcfg['test']=test
        cellcfg['force']=force
        cellcfg['cellbrightp']=f"{cellcfg['outp']}/cellbright.npy"       
        cellbright=cells[cellbox[2]:cellbox[3],cellbox[0]:cellbox[1]]
                                                              
        if not exists(dirname(cellcfg['cellbrightp'])): 
            makedirs(dirname(cellcfg['cellbrightp']),exist_ok=True)
        np.save(cellcfg['cellbrightp'], cellbright) 
        makedirs(dirname(cellcfg['outp']),exist_ok=True)
        makedirs(dirname(cellcfg['plotp']),exist_ok=True)
        
        cellcfg['cellbrightmaskp']=f"{cellcfg['outp']}/cellbrightmask.npy"
        if filterby_centroid:
            # only one cell per box
            from htsimaging.lib.utils import filter_regions
            cellbrightmask=filter_regions(cellbright.astype(int),prop_type='centroid_x',mn=70,mx=80)==0 ## get the region 
        else:
            cellbrightmask=(cellbright.astype(int))==0
            
        np.save(cellcfg['cellbrightmaskp'], cellbrightmask)
        
        cellframe_type2frames={'cellframes':{'frames':[],'ps':[]},
                               'cellframes_masked':{'frames':[],'ps':[]},
                               'cellframes_masked_substracted':{'frames':[],'ps':[]},
                              }
        for dn in sorted(list(cellframe_type2frames.keys())):
            if not exists(f"{cellcfg['outp']}/{dn}"): 
                makedirs(f"{cellcfg['outp']}/{dn}",exist_ok=True)
        from htsimaging.lib.utils import get_signal_summary_by_roi
        for framei,frame in enumerate(frames):
            cellframe=frame[cellbox[2]:cellbox[3],cellbox[0]:cellbox[1]]
            if test and framei==0:
                from htsimaging.viz.image import image_background
                import matplotlib.pyplot as plt
                
                plt.figure()
                image_background(
                    img=cellframe,
                    cmap='gfp',
                ).set(title="Original image")
                
            cellframe_masked=cellframe.copy()
            cellframe_masked[cellbrightmask]=0 ## background to zero
            if test and framei==0:
                plt.figure()
                image_background(
                    img=cellframe_masked,
                    cmap='gfp',
                ).set(title="Remove the extracellular regions")
            
            ## make the cytoplasm intensity uniform
            cellframe_masked_substracted=cellframe_masked.copy()
            signal_cytoplasm=get_signal_summary_by_roi(
                [cellframe_masked_substracted],
                **params_get_signal_summary_by_roi,
                )
            cellframe_masked_substracted=np.where(
                cellframe_masked_substracted<signal_cytoplasm*scale_signal_cytoplasm,
                signal_cytoplasm,
                cellframe_masked_substracted,
                ) 
            cellframe_masked_substracted[cellbrightmask]=0
            if test and framei==0:
                plt.figure()
                image_background(
                    img=cellframe_masked_substracted,
                    cmap='gfp',
                ).set(title="Cytoplasm intensity smoothening")                      

            cellframe_type2frames['cellframes']['frames'].append(cellframe)
            cellframe_type2frames['cellframes_masked']['frames'].append(cellframe_masked)
            cellframe_type2frames['cellframes_masked_substracted']['frames'].append(cellframe_masked_substracted)
            for dn in sorted(list(cellframe_type2frames.keys())):
                cellframe_type2frames[dn]['ps'].append(f"{cellcfg['outp']}/{dn}/frame{framei:08d}.npy")
                np.save(cellframe_type2frames[dn]['ps'][framei],
                        cellframe_type2frames[dn]['frames'][framei])

        for dn in sorted(list(cellframe_type2frames.keys())):
            cellcfg[dn]=cellframe_type2frames[dn]['ps']
        #gfp min max
        for funn in ['min','max']:
            cellcfg[f'cellgfp{funn}p']=f"{cellcfg['outp']}/cellgfp{funn}.npy"
            frame=getattr(np,f"a{funn}")(cellframe_type2frames['cellframes_masked_substracted']['frames'],axis=0)
            np.save(cellcfg[f'cellgfp{funn}p'], frame)
        del frame

        cellcfg['signal_cytoplasm']=get_signal_summary_by_roi(cellframe_type2frames['cellframes_masked']['frames'],
                                                              **params_get_signal_summary_by_roi)

        to_dict(cellcfg,cellcfg['cfgp'])
    else:
        cellcfg=read_dict(cfgp)
    return cellcfg


#!/usr/bin/env python
"""Visualizations."""

import matplotlib.pyplot as plt
import pandas as pd
import numpy as np

def plot_properties_cell(
    cellcfg,
    df2,
    cols_colorby,
    colx='x',
    coly='y',
    ):
    """Plot properties of the cell.

    Args:
        cellcfg (_type_): config of a cell.
        df2 (_type_): input dataframe.
        cols_colorby (_type_): columns to color by.
        colx (str, optional): column with the x values. Defaults to 'x'.
        coly (str, optional): column with the y values. Defaults to 'y'.
    """
    from roux.lib.stat.norm import rescale
    from roux.lib.plot.contour import plot_contourf
    ncols=4
    nrows=int(np.ceil(len(cols_colorby)/4))
    fig,axes=plt.subplots(nrows,ncols,
#                           sharex=True,sharey=True,
                          figsize=[ncols*4,nrows*4]
                         )
#     metric_type='max'
    for axi,(colorby,ax) in enumerate(zip(cols_colorby,np.ravel(axes))):
        fig,ax=plot_contourf(df2[colx],df2[coly],
                      rescale(df2[colorby]),
                     ax=ax,
                     fig=fig,
                      cbar=True if ((axi+1) % 4)==0 else False,
                     params_contourf={'cmap':'binary','vmin':0,'vmax':1,'corner_mask':False},)
        ax=df2.plot.scatter(x=colx,y=coly,color='green',
                         s=10,
                         alpha=0.5,
                         ax=ax)   
        ax.contour(np.load(cellcfg['cellbrightp']), [0.5], linewidths=1, linestyles='dashed',colors='cyan')    
        ax.set_title(colorby)
        ax.invert_yaxis()
        ax.axis('equal')
#         ax.set_ylim(ax.get_ylim()[::-1])
#         ax.set_axis_off()
    plt.tight_layout()

def image_locate_particles(
    df1: pd.DataFrame,
    frame,
    img_region,
    annotate_particles: str=False,
    fig=None,
    ax: plt.Axes=None,
    ) -> plt.Axes:
    """Plot image with particles.

    Args:
        df1 (pd.DataFrame): input dataframe.
        frame (_type_): image frame.
        img_region (_type_): regions in the image.
        annotate_particles (str, optional): annotate the paticles or not. Defaults to False.
        fig (_type_, optional): figure object. Defaults to None.
        ax (plt.Axes, optional): subplot object. Defaults to None.

    Returns:
        plt.Axes: _description_
    """
    import trackpy as tp
    fig=plt.figure(figsize=[20,20]) if fig is None else fig
    ax=plt.subplot(111) if ax is None else ax
    from htsimaging.viz.image import image_background
    ax=image_background(img_region=img_region,img=frame,ax=ax)
    ax=tp.annotate(df1, frame,ax=ax)
    if annotate_particles:
        _=df1.apply(lambda x:ax.text(x['x'],x['y'],int(x['particle']),color='lime'),axis=1)
#     ax.grid(False)
    return ax

def image_trajectories(
    dtraj: pd.DataFrame,
    img_gfp=None,
    img_bright=None,
    label: bool=True,
    fig=None,
    ax: plt.Axes=None,
    ) -> plt.Axes:
    """Plot trajectories.

    Args:
        dtraj (pd.DataFrame): input dataframe with the trajectories.
        img_gfp (_type_, optional): channel image e.g. GFP. Defaults to None.
        img_bright (_type_, optional): segmentation image e.g. bright field. Defaults to None.
        label (bool, optional): label. Defaults to True.
        fig (_type_, optional): figure object. Defaults to None.
        ax (plt.Axes, optional): subplot object. Defaults to None.

    Returns:
        plt.Axes: subplot
    """
    import trackpy as tp    
    fig=plt.figure(figsize=[20,20]) if fig is None else fig
    ax=plt.subplot(111) if ax is None else ax
    from htsimaging.viz.image import image_background
    ax=image_background(
        img_region=img_bright,img=img_gfp,
        cmap='gfp',
        ax=ax)
    ax = tp.plot_traj(dtraj,label=False,ax=ax,lw=2,plot_style={'color':'k'})
    dtrajagg=dtraj.groupby('particle').agg({c:np.median for c in ['x','y']}).reset_index()
#     dtrajagg.columns=coltuples2str(dtrajagg.columns)
#     print(dtrajagg.iloc[0,:])
    if label:
        dtrajagg.reset_index().apply(lambda x: ax.text(x['x'],x['y'],int(x['particle']),
                                    color='magenta'),
                  axis=1)
    ax.grid(False)
    return ax

def plot_moving_particles(
    t_cor: pd.DataFrame,
    img_bright=None,
    frame=None,
    framei: int=0,
    particle2color=None,
    test: bool=False,
    outd: str=None,
    ):
    """Plot moving particles.

    Args:
        t_cor (pd.DataFrame): input table
        img_bright (_type_, optional): segmentation raw image (e.g. bright field). Defaults to None.
        frame (_type_, optional): image frame. Defaults to None.
        framei (int, optional): image frame index. Defaults to 0.
        particle2color (_type_, optional): particle-wise colors. Defaults to None.
        test (bool, optional): test-mode. Defaults to False.
        outd (str, optional): path to the output directory. Defaults to None.
    """
    plotp=f'{outd}/{framei:03d}.jpeg'
    plt.figure(figsize=[5,5])
    ax=plt.subplot(111)
    if not ((img_bright is None) and (frame is None)):
        from htsimaging.viz.image import image_background
        ax=image_background(img_region=img_bright,img=frame,ax=ax)
    ax.text(ax.get_xlim()[1],ax.get_ylim()[1],
            f"frame#{framei:03d}",ha='right',
            va='bottom',
            size='20',
            color='k')
    for p in t_cor.loc[(t_cor['frame']==framei),'particle'].unique():
        framei_min=t_cor['frame min'].unique()[0]
        df=t_cor.loc[((t_cor['particle']==p) \
                      & (t_cor['frame'].isin(range(framei_min,framei+1)))\
                      & (t_cor['x'].between(0,frame.shape[0]))\
                      & (t_cor['y'].between(0,frame.shape[1]))),:].sort_values(by=['frame','particle'])
        if len(df)!=0:
            if framei!=0:    
                df.plot.line(x='x',y='y',lw=4,
                    c='limegreen' if particle2color is None else particle2color[p],
                    legend=False,ax=ax)
            else:
                df.plot.scatter(x='x',y='y',s=4,
                    c='limegreen' if particle2color is None else particle2color[p],
                    legend=False,ax=ax)
            if test:
                d=df.head(1).T.to_dict()
                d=d[list(d.keys())[0]]
                ax.text(x=d['x'],y=d['y'],s=int(d['particle']),
                    color='magenta',
                    )
    ax.set_xlim(0,frame.shape[1])
    if not test:
        plt.axis('off')
        makedirs(dirname(plotp),exist_ok=True)
        plt.savefig(plotp)
        plt.clf();plt.close();



#!/usr/bin/env python
"""I/O"""

from os.path import exists
from glob import glob

import numpy as np
import pandas as pd 

def to_frames(
    input_path: str,
    channeli=None,
    ):
    """
    Convert to frames.

    Args:
        input_path (str): path to the raw data.

    Returns:
        list: list of frames.
    """
    if input_path.endswith('nd2'):
        frames=pims.ND2_Reader(input_path)
    elif input_path.endswith('mp4'):
        import imageio
        vid = imageio.get_reader(input_path,  'ffmpeg')
        frames=[frame for frame in vid.iter_data()]
        if not channeli is None:
            # get a channel 
            frames=[frame[:,:,1] for frame in vid.iter_data()]
    if input_path.endswith('nd2'):
        if len(np.shape(frames))==4:
            frames = map(lambda x: x.mean(axis=0), frames)
    # else:
    #     frames.default_coords['c'] = 1
    #     frames.bundle_axes='yx'
    #     frames.iter_axes = 't'
    return frames

def expt_dh2expt_info(
    expt_dh: str,
    ) -> pd.DataFrame:
    """
    Make configuration using the directory structure for an experiment.

    Args:
        expt_dh (str): str

    Returns:
        pd.DataFrame: output dataframe.
    """
    if exists(expt_dh+"info"):
        expt_info=pd.read_csv(expt_dh+"info")
    elif exists(expt_dh+"info.csv"):
        expt_info=pd.read_csv(expt_dh+"info.csv")
    cols_del=["dh","fn_lead"]
    for col in expt_info.columns.tolist():
        if "Unnamed" in col:
            cols_del.append(col)
    for col in cols_del:
        if col in expt_info.columns.tolist():
            del expt_info[col]    
    expt_info2=expt_info.drop(["smpn"],axis=1).T
    expt_info2.columns=expt_info.loc[:,"smpn"]
    
    # print expt_info2
    for col in expt_info2.columns.tolist():
        for i in range(len([i for i in expt_info2.index if 'replicate' in i])):
            if not pd.isnull(expt_info2.loc["replicate %d" % (i+1),col]):
                expt_info2.loc[("replicate %d" % (i+1)),col]=['replicate %d' % (i+1),expt_info2.loc[("replicate %d" % (i+1)),col]]
            else:
                expt_info2.loc[("replicate %d" % (i+1)),col]=[np.nan,np.nan]
    expt_info2=expt_info2.reset_index()
    del expt_info2["index"]
    return expt_info2

def createinfo(
    expt_dh: str,
    ) -> str:
    """
    Create information file.

    Args:
        expt_dh (str): path to the directory containing the raw data.

    Returns: 
        path to the file containing the metadata. 
    """
    info=pd.read_csv(expt_dh+"info")
    for i in range(len(info)):
        reps=glob("%s/%s*" % (info.loc[i,"dh"],info.loc[i,"fn_lead"]))
        for repi,rep in enumerate(reps):
            info.loc[i,"replicate %d" % (repi+1)]=rep
    #     break
    output_path=expt_dh+"info"
    info.to_csv(output_path)
    return output_path


#!/usr/bin/env python
"""GLobally used variables."""
from matplotlib import colors
import numpy as np

r=np.linspace(0,0.14,10)
g=np.linspace(0,1,10)
b=np.linspace(0,0,10)
cmap_gfp_list=[l for l in zip(r,g,b)]
cmap_gfp = colors.LinearSegmentedColormap.from_list("gfp",cmap_gfp_list)


#!/usr/bin/env python
"""To make the video of the timelapse images."""

import numpy as np
import pandas as pd

import matplotlib.pyplot as plt

def make_gif(
    cellcfg=None,
    frames: list=None,
    t_cor: pd.DataFrame=None,
    img_bright=None,
    outd: str=None,
    particle2color: dict=None,
    test: bool=False,
    force: bool=False,
    ):
    """
    Make a .gif file out of frames. 
    """
    if not cellcfg is None:
        if frames is None:
            frames=[np.load(p) for p in sorted(cellcfg['cellframes'])]
        if t_cor is None:
            t_cor=read_table(f"{cellcfg['outp']}/d2filter_stubs.tsv")
        if img_bright is None:
            img_bright=np.load(cellcfg['cellbrightp'])
        if outd is None:    
            outd=f"{cellcfg['outp']}/vid"
        if not particle2color is None:
            if exists(f"{cellcfg['outp']}/d4distance.tsv"):
                df1=read_table(f"{cellcfg['outp']}/d4distance.tsv").drop_duplicates(subset=['particle'])
                from roux.stat.norm import rescale
                from roux.viz.colors import saturate_color
                df1['distance effective from centroid']=rescale(df1['distance effective from centroid'])
                df1.loc[(df1['distance effective from centroid']<=0.5),'distance effective from centroid']=0.5
                df1['color']=df1['distance effective from centroid'].apply(lambda x: saturate_color('limegreen',x))
                particle2color=df1.set_index('particle')['color'].to_dict()
            else:
                logging.warning('distance file (d4distance.tsv) not found')
    makedirs(outd,exist_ok=True)
    gifp=f"{dirname(outd)}/{basenamenoext(outd)}.gif"

    if exists(gifp) and not force:
        return
    if not 'frame min' in t_cor:
        df=t_cor.groupby('particle').agg({'frame':[min,max]})
        df.columns=coltuples2str(df.columns)
        t_cor=t_cor.merge(df,left_on='particle',right_index=True)
    t_cor=t_cor.sort_values(['frame','particle'])
    for framei,frame in enumerate(frames):
        plot_moving_particles(t_cor,img_bright=img_bright,frame=frame,
                            framei=framei,particle2color=particle2color,
                             outd=outd,
                             test=False)
    if not test:    
        plt.close('all')
        com=f"convert -delay 10 -loop 0 {outd}/*.jpeg {gifp}"
        from roux.lib.sys import runbashcmd
        runbashcmd(com)
    return gifp



#!/usr/bin/env python
"""Statistical analysis of the bleach-chase"""
import pandas as pd

def exp1(x,lag,amplitude,rate):
    """
    One-term exponential.
    
    Parameters:
        x (list): input vector
        lag (float): lag   
        amplitude (float): amplitude
        rate (float): rate
    """
    return lag+(amplitude*(1-np.exp(-rate*x)))
def get_scores(
    df: pd.DataFrame,
    ) -> pd.DataFrame:
    """
    Calculate the rates other parameters.
    
    Parameters:
        df (pd.DataFrame): input table.
    
    Returns:
        pd.DataFrame: table with the parameters.
    """
    popt, pcov = curve_fit(
        exp1, 
        xdata=df['time'], 
        ydata=df['abundance rescaled'],
        p0=[0,1,0.01],
        bounds=([0,0.9,0],[0.1,1.1,1]),
    )

    perr = np.sqrt(np.diag(pcov))
    df_=pd.DataFrame({'score':popt,'score SD':perr,})
    df_.index=['lag','amplitude','rate']
    df_.index.name='variable'
    return df_



#!/usr/bin/env python
"""Processing of the segmented regions."""

import logging
import numpy as np
from skimage import measure

from htsimaging.lib.utils import filter_regions
from htsimaging.lib.io import read_image

def segmentation2cells(
    # imp: str,
    imsegp: str,
    kind : str = 'yeast',
    fiterby_border_thickness: int=None,
    magnification: int=100,
    test: bool=False,
    **kws: dict,
    ) -> list:
    """
    Segment the image to the single cells.

    Args:
        imsegp (str): _description_
        fiterby_border_thickness (int, optional): _description_. Defaults to 100.
        magnification (int, optional): _description_. Defaults to 100.
        plotp (str, optional): _description_. Defaults to None.

    Returns:
        list: _description_

    Examples:
        1. Parameters:
            prop_type='area',mn=100,mx=8000
            at 1.5X
            prop_type='area',mn=1500,mx=12000
    """
    # im=read_image(imp)
    imseg=read_image(imsegp)
        
    regions=measure.label(imseg)
#     fiterby_border_thickness,im.shape[1]+fiterby_border_thickness
    if kind=='yeast':
        if test: logging.info(f"number of regions = {len(np.unique(regions))}")
        regions=filter_regions(regions,imseg,prop_type='area',
                               mn=10*magnification,
                               mx=100000*magnification,
                               plotp=plotp,
                               **kws,
                              )
        if test: logging.info(f"number of regions = {len(np.unique(regions))}")
        regions=filter_regions(regions,imseg,prop_type='eccentricity',mn=0,mx=0.8,test=test,
                               **kws,
                              )
    if test: logging.info(f"number of regions = {len(np.unique(regions))}")
    if not fiterby_border_thickness is None:
        regions=filter_regions(regions,imseg,prop_type='centroid_x',
                               mn=fiterby_border_thickness,mx=imseg.shape[0]+fiterby_border_thickness,test=test,
                               **kws,
                              )
        if test: logging.info(f"number of regions = {len(np.unique(regions))}")
        regions=filter_regions(regions,imseg,prop_type='centroid_y',
                               mn=fiterby_border_thickness,mx=imseg.shape[1]+fiterby_border_thickness,test=test,
                               **kws,
                              )
        if test: logging.info(f"number of regions = {len(np.unique(regions))}")
    if len(np.unique(regions))==1:
        logging.error("the output does not contain any regions")
    return regions               
               
def get_cellboxes(
    regions: list,
    cellbox_width: int=150,
    test: bool=False,
    ) -> list:
    """
    Get the bounding boxes of the cells.

    Args:
        regions (list): regions.
        cellbox_width (int, optional): width of the bounding box of the cell. Defaults to 150.
        test (bool, optional): test-mode. Defaults to False.
    Returns:
        list: list of the bounding boxes for cells. 
    """
    import matplotlib.patches as mpatches
    if test:
        import matplotlib.pyplot as plt
        fig, ax = plt.subplots(figsize=(10, 6))
        plt.imshow(regions,cmap='binary')
    cellboxes=[]
    for regioni,region in enumerate(measure.regionprops(regions.astype(int))):
        box_xmnxmxymnymx=[
            region.centroid[1]-(cellbox_width*0.5),
            region.centroid[1]+(cellbox_width*0.5),
            region.centroid[0]-(cellbox_width*0.5),
            region.centroid[0]+(cellbox_width*0.5),
                         ]
        cellboxes.append([int(i) for i in box_xmnxmxymnymx])
#         celli2props[regioni+1]=region.area
        if test:
            rect = mpatches.Rectangle([box_xmnxmxymnymx[0],box_xmnxmxymnymx[2]], cellbox_width, cellbox_width,
                                      fill=False, edgecolor='red', linewidth=2)
            ax.text(region.centroid[1],region.centroid[0],f"{regioni+1:d}",color='g')
#             ax.text(region.centroid[1],region.centroid[0],f"{region.extent:.2f}",color='g')
            ax.add_patch(rect)
    return cellboxes

def arr_list2regions(
    arr_list: list,
    time_increment: int,
    ) -> tuple:
    """Parameterized cell-segmentation for the time lapse images. 

    Args:
        arr_list (list): frames of images.
        time_increment (int): time interval.

    Returns:
        tuple: regions and table with intensities.
        
    """
    pre_bleach=arr_list[0]
    denoised=restoration.denoise_bilateral(pre_bleach.astype('uint16'), sigma_range=0.01, sigma_spatial=15)
    smoothened = filters.median(denoised,np.ones((4,4)))
    markers = np.zeros(smoothened.shape, dtype=np.uint)
    markers[smoothened < filters.threshold_otsu(smoothened)] = 1
    markers[smoothened > filters.threshold_otsu(smoothened)] = 2
    labels = random_walker(smoothened, markers, beta=10, mode='bf')
    regions= measure.label(labels)
    props = measure.regionprops(regions,arr_list[0]) #ref
    regions_areas=np.array([prop.area  for prop in props])
    regions_lbls =np.array([prop.label for prop in props])
    regions_means=np.array([prop.mean_intensity  for prop in props])
    regions_index_large=np.where((regions_areas<10000) & (regions_areas>200) & (regions_means>2000))[0]
    regions_lbls_large=regions_lbls[regions_index_large] 
    regions_large=np.zeros((regions.shape), dtype=bool)
    for i in regions_lbls_large:
        booli = (regions == i) 
        regions_large=np.logical_or(regions_large,booli)
    kins_mean=pd.DataFrame(columns=regions_lbls_large, index=range(len(arr_list)))
    for i in range(len(arr_list)):
        props = measure.regionprops(regions,arr_list[i])
        means=np.array([prop.mean_intensity for prop in props])
        kins_mean.loc[i,:]=means[regions_index_large]
        del props
    kins_mean=kins_mean.loc[:, ~(kins_mean < 3000).any(axis=0)] #stitch
    kins_mean['time']=np.array(range(len(arr_list)))*time_increment #stitch
    return regions_large,kins_mean



#!/usr/bin/env python
"""Statistics."""

import numpy as np
import pandas as pd

import matplotlib.pyplot as plt
from scipy.optimize import leastsq

def power(x, A, B):
    """power law equation."""
    return B*x**A
def power_residuals(p, y, x):
    """Deviations of data from fitted 4PL curve"""
    A,B = p
    err = y-power(x, A, B)
    return err
def power_peval(x, p):
    """Evaluated value at x with current parameters."""
    A,B = p
    return power(x, A, B)

def line(x, m, C):
    """power law equation."""
    return m*x+C
def line_residuals(p, y, x):
    """Deviations of data from fitted 4PL curve"""
    m,C = p
    err = y-line(x, m,C)
    return err
def line_peval(x, p):
    """Evaluated value at x with current parameters."""
    m,C = p
    return line(x, m,C)

def logistic4(x, A, B, C, D):
    return ((A-D)/(1.0+((x/C)**B))) + D
def logistic4_residuals(p, y, x):
    """Deviations of data from fitted 4PL curve"""
    A,B,C,D = p
    err = y-logistic4(x, A, B, C, D)
    return err
def logistic4_peval(x, p):
    """Evaluated value at x with current parameters."""
    A,B,C,D = p
    return logistic4(x, A, B, C, D)

def logistic5(x, A, B, C, D, E):
    return  D+(A-D)/((1+(x/C)**B)**E)
def logistic5_residuals(p, y, x):
    """Deviations of data from fitted 4PL curve"""
    A,B,C,D,E = p
    err = y-logistic5(x, A, B, C, D, E)
    return err
def logistic5_peval(x, p):
    """Evaluated value at x with current parameters."""
    A,B,C,D,E = p
    return logistic5(x, A, B, C, D, E)

def fit_power(
    x,
    y,
    p0: list=[0, 1],
    plot: bool=False,
    ):
    """
    Fit power law.

    Args:
        x (vector-like): x values.
        y (vector-like): y values.
        p0 (list, optional): starting parameters. Defaults to [0, 1].
        plot (bool, optional): to plot or not. Defaults to False.

    Returns:
        tuple: outputs.
    """
    plsq,cov,infodict,mesg,ier= leastsq(power_residuals, p0, args=(y, x),full_output=1)
    ss_err=(infodict['fvec']**2).sum()
    ss_tot=((y-y.mean())**2).sum()
    rsquared=1-(ss_err/ss_tot)
    # if rsquared<0.85:
        # print "poor fit %d" % rsquared
    if plot==True:
        plt.plot(x,y,'o')  
        plt.plot(x,power_peval(x,plsq)) 
    return plsq[0], plsq[1],rsquared,power_peval(np.max(x),plsq)

def fit_line(
    x,
    y,
    p0: list=[0, 1],
    plot: bool=False,
    ) -> tuple:
    """
    Fit power law.

    Args:
        x (vector-like): x values.
        y (vector-like): y values.
        p0 (list, optional): _description_. Defaults to [0, 1].
        plot (bool, optional): _description_. Defaults to False.

    Returns:
        tuple: outputs
    """
    plsq,cov,infodict,mesg,ier= leastsq(line_residuals, p0, args=(y, x),full_output=1)
    ss_err=(infodict['fvec']**2).sum()
    ss_tot=((y-y.mean())**2).sum()
    rsquared=1-(ss_err/ss_tot)
    # if rsquared<0.85:
    #     print "poor fit %d" % rsquared
    if plot==True:
        plt.plot(x,y,'o')  
        plt.plot(x,line_peval(x,plsq)) 
    return plsq[0], plsq[1],rsquared

def get_slope(
    df: pd.DataFrame,
    ds: pd.Series,
    ) -> float:
    """
    Get slope for a section of the line.

    Args:
        df (pd.DataFrame): input dataframe.
        ds (pd.Series): section of the line.

    Returns:
        float: slope.
    """
    from scipy.stats import linregress
    return linregress(df.loc[ds.index,'frame'],df.loc[ds.index,'distance effective from centroid per frame']).slope

def get_inflection_point(
    df: pd.DataFrame,
    threshold_slope: float=0.25,
    ) -> pd.DataFrame:
    """
    Get inflation point.

    Args:
        df (pd.DataFrame): input dataframe.
        threshold_slope (float, optional): threshold on the slope. Defaults to 0.25.

    Returns:
        pd.DataFrame: output dataframe.
    """
    label=df.name
    df['slope distance effective from centroid versus frame']=df.rolling(6)['y'].apply(lambda x: get_slope(df,x),raw=False)
    if any(df['slope distance effective from centroid versus frame']>threshold_slope) and not all(df['slope distance effective from centroid versus frame']>threshold_slope):
        inflection_point=df.set_index('frame')['slope distance effective from centroid versus frame'].idxmax()
    else:
        inflection_point=None
    df['inflection point']=inflection_point
    return df


## test examples using papermill
import logging
from pathlib import Path
import os
from os.path import exists

from testbook import testbook
import papermill as pm

print(">>>>>>>>>>>>>>>>> current dir:"+os.getcwd())
if not os.getcwd().endswith("/examples"):
    os.chdir('./examples/')
    print(">>>>>>>>>>>>>>>>> current dir:"+os.getcwd())

@testbook("_make_demo_data.ipynb", execute=True)
def test_make_demo_data(tb):
    pass # execute only because tests are present in the notebook itself
    return

# parameters
input_dir_path='./inputs/'

def test_protein_abundance_and_normalization(
    input_path='./protein_abundance_and_normalization.ipynb',
    ):
    output_dir_path= './outputs/' + Path(input_path).stem + '/'
    os.makedirs(output_dir_path,exist_ok=True)

    parameters=dict(
    ## parameters
        input_path=f'{input_dir_path}/image_intensity_cells.npy',
        segmented_image_path=f'{input_dir_path}/image_segmented_cells.npy',
        output_path=f'{output_dir_path}/01_gfpby_cell.tsv',
        misaligned_fraction_max=0.9,
        edge_off=20,
    )
    
    logging.info(parameters)
    Path(output_dir_path).mkdir(exist_ok=True)

    _=pm.execute_notebook(
        input_path=input_path,
        output_path='./outputs/' + Path(input_path).name,
        parameters=parameters,
        # report_mode=True,
        kernel_name='test',
        execution_timeout=360,
    )
    assert exists(parameters['output_path'])
    
def test_protein_abundance_by_marker_location(
    input_path='./protein_abundance_by_marker_location.ipynb',
    ):
    output_dir_path= './outputs/' + Path(input_path).stem + '/'
    os.makedirs(output_dir_path,exist_ok=True)

    parameters=dict(
    ## parameters
        ## parameters
        input_path=f'{input_dir_path}/image_marker_cells.npy', ## path to the image from channel marking a subcellular localization
        output_path=f'{output_dir_path}/01_gfpby_cell_and_marker.tsv',

        image_intensity_path=f'{input_dir_path}/image_intensity_cells.npy', ## path to the image with channel which is to be used to caluculate the abundance
        regions_path=f'{input_dir_path}/image_segmented_cells.npy', ## segmented regions (dtype: bool)
        marker_intensity_min_quantile=0.975,
        pixels_per_cell_min=100,
        non_marker_intensity_quantile_off=0.02,

        force=False,
        test=True,
    )

    logging.info(parameters)
    Path(output_dir_path).mkdir(exist_ok=True)

    _=pm.execute_notebook(
        input_path=input_path,
        output_path='./outputs/' + Path(input_path).name,
        parameters=parameters,
        # report_mode=True,
        kernel_name='test',
        execution_timeout=360,
    )
    assert exists(parameters['output_path'])



"""
Tests the examples.
Notes: The paths are relative to '../'
"""
import logging
import os
from testbook import testbook

print(">>>>>>>>>>>>>>>>> current dir:"+os.getcwd())
if not os.getcwd().endswith("/examples"):
    os.chdir('./examples/')
    print(">>>>>>>>>>>>>>>>> current dir:"+os.getcwd())

@testbook("viz_image.ipynb", execute=True)
def test_viz_image(tb):
    pass # execute only because tests are present in the notebook itself
    return



#!/usr/bin/env python
"""Statistical analysis of the single particle tracking."""

import logging
logging.basicConfig(format='[%(asctime)s] %(levelname)s\tfrom %(filename)s in %(funcName)s(..): %(message)s',level=logging.INFO) # 

from os import makedirs
from os.path import exists, dirname

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from roux.lib.df import flatten_columns
from roux.lib.io import read_dict, to_dict, to_table

import trackpy as tp
tp.ignore_logging()

from scipy.spatial import distance

def test_locate_particles(
    cellcfg: dict,
    params_locate: dict,
    frame=None,
    force: bool=False,
    test: bool=False,
    ) -> bool:
    """
    Test locating of the particles.

    Args:
        cellcfg (dict): the cell level configuration. 
        params_locate (dict): parameters provided for the location.
        frame (np.array, optional): image frame. Defaults to None.
        force (bool, optional): over-write the outputs. Defaults to False.
        test (bool, optional): test mode. Defaults to False.

    Returns:
        bool
    """
    dlocate_testp=f"{cellcfg['outp']}/dlocate_test.tsv"
    if exists(dlocate_testp) and not force and not test:
        return True
    cellgfpmax=np.load(cellcfg['cellgfpmaxp'])
    frame=cellgfpmax if frame is None else frame
    df1 = tp.locate(frame, 
                    **params_locate)
    df1['particle']=df1.index
    if not test:
        to_table(df1,dlocate_testp)
    logging.info(f"particles detected ={len(df1)}") if not test else print(f"particles detected ={len(df1)}")
    if len(df1)==0:
        return False
    # plot dist signal of the detected particles
    fig=plt.figure()
    cellgfpmin=np.load(cellcfg['cellgfpminp'])
    from htsimaging.viz.stat import dist_signal
    ax=plt.subplot()
#     dist_signal(np.unique(cellgfpmin)[2:],
#                 params_hist={'bins':20,'label':'gfp min',
#                              'density':True,'color':'k'},ax=ax)
#     dist_signal(np.unique(cellgfpmax)[2:],
#                 params_hist={'bins':20,'label':'gfp max',
#                              'density':True,'color':'green'},ax=ax)
    dist_signal(np.unique(frame)[2:],
                params_hist={'bins':20,'label':'frame',
                             'density':True,'color':'lime'},ax=ax)        
    dist_signal(df1['signal'],
                threshold=np.unique(frame)[1],label_threshold='signal_cytoplasm',
                params_hist={'bins':20,'label':f'particles\n(total ={len(df1)})',
                             'density':True,'color':'r'},ax=ax)
    if not test:
        ax.set(title='QC: Particle detection')
        plt.savefig(f"{cellcfg['plotp']}/dist_signal_locate_particles.png")
    # plot image of the detected particles
    from htsimaging.endocytosis.viz import image_locate_particles
    ax=image_locate_particles(df1,
                           frame=frame,
                           img_region=np.load(cellcfg['cellbrightp']),
                           annotate_particles=False)
    _=ax.text(0,1,'\n'.join([f"{k}:{params_locate[k]}" for k in params_locate]),
              va='top',color='lime')
    if not test:
        ax.set(title='QC: Particle detection')
        plt.savefig(f"{cellcfg['plotp']}/image_locate_particles.png")  
    return True

def to_msd(
    frames: list,
    coff_intesity_perc : float = 75,
    diameter=11,
    cores : int = 4,
    test : bool = False,
    ) -> tuple:
    """
    MSD from the nd file.

    Args:
        frames (str): 2D frames.
        
        
    Returns:
        tuple: outputs.
    """
    logging.info('number of frames = %d' % len(frames))
    assert len(np.shape(frames))<4, f"{np.shape(frames)} perhaps use average_z(frames)"
    threshold=np.percentile(np.hstack(tuple(frames)),coff_intesity_perc)
    f_batch = tp.batch(frames,diameter=diameter,threshold=threshold,processes=cores)
    if test:
        plt.figure()
        ax=tp.annotate(f_batch.query("`frame` == 0"), frames[0])
        ax.set(title="Test: Detection of the particles")
    t = tp.link_df(f_batch, search_range=diameter, memory=3)
    t_flt = tp.filter_stubs(t, 3*int(len(frames)/4))
    try:
        d = tp.compute_drift(t_flt)
        t_cor = tp.subtract_drift(t_flt, d)
    except:
        t_cor=t_flt
        logging.info("drift correction excepted")
    imsd=tp.imsd(t_cor,0.1,0.2, max_lagtime=100, statistic='msd')
    emsd=tp.emsd(t_cor,0.1,0.2, max_lagtime=100)
    return imsd,emsd    
        
def trim_returns(
    df1: pd.DataFrame,
    ) -> pd.DataFrame:
    """
    Trim images.

    Args:
        df1 (pd.DataFrame): input dataframe.

    Returns:
        pd.DataFrame: output dataframe.
    """
    from htsimaging.lib.stat import get_inflection_point
    from htsimaging.spt.viz import plot_trajectories_stats
    # get inflection point if any
    df1=df1.groupby('particle').apply(lambda x: get_inflection_point(x))
    df2=df1.groupby('particle',as_index=False).apply(lambda df: df.set_index('frame').sort_index(axis=0).loc[df['inflection point'].unique()[0]:,:].reset_index())


    fig=plt.figure(figsize=[20,10])
    # plot before filtering
    ax=plt.subplot()
    df1.loc[df1['inflection point'].isnull(),:].groupby('particle').apply(lambda x: plot_trajectories_stats(x,
                                                            coly='distance effective from centroid per frame',
                                                            colx='frame',
                                                           fig=fig,ax=ax,
                                                           rescalex=False,
                                                          params_plot={'color':'lime','alpha':0.5}))
    df1.loc[~df1['inflection point'].isnull(),:].groupby('particle').apply(lambda x: plot_trajectories_stats(x,
                                                             coly='distance effective from centroid per frame',
                                                             colx='frame',
                                                           fig=fig,ax=ax,
                                                           rescalex=False,
                                                            axvlinex=x['inflection point'].unique()[0],
                                                          params_plot={'color':'r','alpha':0.75}))

    df2.loc[~df2['inflection point'].isnull(),:].groupby('particle').apply(lambda x: plot_trajectories_stats(x,
                                                             coly='distance effective from centroid per frame',
                                                             colx='frame',
                                                           fig=fig,ax=ax,
                                                           rescalex=False,
                                                            axvlinex=x['inflection point'].unique()[0],
                                                          params_plot={'color':'g','alpha':1}))
    ax.get_legend().remove()
    ax.set_ylim(df1['distance effective from centroid per frame'].min(),df1['distance effective from centroid per frame'].max())               
    ax.set_xlim(df1['frame'].min(),df1['frame'].max())
    ax.set(title='Returns trimmed')
    return df2
                
def fill_frame_jumps(
    df1: pd.DataFrame,
    jump_length,
    ) -> pd.DataFrame:
    """
    Fill the frame jumps.

    Args:
        df1 (pd.DataFrame): input dataframe.
        jump_length (_type_): length of the jump.

    Returns:
        pd.DataFrame: output dataframe.
    """
    df1=df1.sort_values(by=['particle','frame'])
    df1['frame delta']=df1['frame']-([df1['frame'].tolist()[0]-1]+df1['frame'].tolist()[:-1])
    df=df1.loc[(df1['frame delta']==jump_length),:]
    df.loc[df.index,'frame']=df['frame']-1
    df1=pd.concat({'not':df1,'fill':df},axis=0)
    df1.index.name='frame fill or not'
    df1=df1.reset_index()
    df1=df1.sort_values(by=['particle','frame'])
    df1['frame delta']=df1['frame']-([df1['frame'].tolist()[0]-1]+df1['frame'].tolist()[:-1])
    logging.warning(sum((df1['frame']-([df1['frame'].tolist()[0]-1]+df1['frame'].tolist()[:-1]))==jump_length)==0)
    df1.index=range(len(df1))
    return df1                
            
def cellcfg2distances(
    cellcfg: dict,
    params: dict,
    subtract_drift: bool=False,
    test: bool=False,
    force: bool=False,
    ):
    """
    Calculate distances from cell configuration.

    Args:
        cellcfg (dict): configuration
        params (_type_, optional): parameters. Defaults to { 'locate':{'diameter':11, # round to odd number 'noise_size':1, 'separation':15, 'threshold':4000, 'preprocess':True, 'invert':False, 'max_iterations':50, 'percentile':0, 'engine':'numba', }, 'link_df':{ 'search_range':5, 'memory':1, 'link_strategy':'drop',}, 'filter_stubs':{'threshold':4}, 'get_distance_from_centroid':{'center':[75,75]}, }.
        force (bool, optional): over-write the outputs. Defaults to False.

    """
    params['locate']['separation']=params['locate']['diameter']*1
    params['locate']['threshold']=cellcfg['signal_cytoplasm']*0.5
    params['link_df']['search_range']=params['locate']['diameter']*0.33
            
    to_dict(params,f"{cellcfg['outp']}/params.yml")
    
    if not test_locate_particles(cellcfg,params['locate'],force=force,test=False):
        print(cellcfg['cfgp'])
        return 
    # get trajectories
    steps=['locate','link_df','filter_stubs','filter_returns','subtract_drift','distance']
    dn2dp={s:f"{cellcfg['outp']}/d{si}{s}.tsv" for si,s in enumerate(steps)}
    dn2plotp_suffix={s:f"{si}{s}.png" for si,s in enumerate(steps)}
    steps_done=[k for k in dn2dp if exists(dn2dp[k])]
    
    if ('distance' in steps_done) and not force:
        print(cellcfg['cfgp'])
        return
           
    from htsimaging.endocytosis.viz import image_trajectories
    from htsimaging.spt.stat import get_distance_from_centroid
           
    img_gfp=np.load(cellcfg['cellgfpmaxp'])
    img_bright=np.load(cellcfg['cellbrightp'])
    
    dn2df={}
    dn2df['locate']=tp.batch([np.load(p) for p in sorted(cellcfg['cellframes_masked_substracted'])],
                             **params['locate'])
    if len(dn2df['locate'])==0:
        return
    dn2df['locate']['frame']=dn2df['locate']['frame'].astype(np.integer)
    dn2df['link_df']=tp.link_df(dn2df['locate'], **params['link_df'])
#     if params['link_df']['memory']!=0:
    dn2df['link_df']=fill_frame_jumps(dn2df['link_df'],
                      jump_length=2 if params['link_df']['memory']==0 else params['link_df']['memory']+1)
#     to_table(dn2df['link_df'],'test.tsv')
#     to_table(dn2df['link_df'],dn2dp['link_df'])
    image_trajectories(
        dtraj=dn2df['link_df'], 
        img_gfp=img_gfp, 
        img_bright=img_bright,
        fig=None,
        ax=None,
        ).set(title='Particles linked')
    makedirs(cellcfg['plotp'],exist_ok=True)
    plt.savefig(f"{cellcfg['plotp']}/image_trajectories_{dn2plotp_suffix['link_df']}")
#     to_table(dn2df['link_df'],dn2dp['link_df'])

    
    dn2df['filter_stubs']=tp.filter_stubs(dn2df['link_df'], threshold=params['filter_stubs']['threshold'])
    dn2df['filter_stubs'].index.name='index'
    dn2df['filter_stubs'].index=range(len(dn2df['filter_stubs']))
    if len(dn2df['filter_stubs'])==0:
        to_table(dn2df['filter_stubs'],dn2dp['distance'])
        print(cellcfg['cfgp'])
        return 
    image_trajectories(dtraj=dn2df['filter_stubs'], 
                       img_gfp=img_gfp, 
                       img_bright=img_bright, fig=None, ax=None).set(title='Stubs removed')
    plt.savefig(f"{cellcfg['plotp']}/image_trajectories_{dn2plotp_suffix['filter_stubs']}")
    
    dn2df['filter_returns']=get_distance_from_centroid(dn2df['filter_stubs'],**params['get_distance_from_centroid'])
    dn2df['filter_returns']=trim_returns(dn2df['filter_returns'])
    plt.savefig(f"{cellcfg['plotp']}/image_trajectories_stats_trimming_{dn2plotp_suffix['filter_returns']}")

    dn2df['filter_returns']=tp.filter_stubs(dn2df['filter_returns'], threshold=params['filter_stubs']['threshold'])
    dn2df['filter_returns'].index.name='index'
    dn2df['filter_returns'].index=range(len(dn2df['filter_returns']))
    if len(dn2df['filter_returns'])==0:
        to_table(dn2df['filter_returns'],dn2dp['distance'])
        print(cellcfg['cfgp'])
        return             
    image_trajectories(dtraj=dn2df['filter_stubs'], 
                       img_gfp=img_gfp, 
                       img_bright=img_bright, fig=None, ax=None).set(title='Filtered returns')
    plt.savefig(f"{cellcfg['plotp']}/image_trajectories_{dn2plotp_suffix['filter_returns']}")
    if subtract_drift:
        d = tp.compute_drift(dn2df['filter_returns'])
        dn2df['subtract_drift'] = tp.subtract_drift(dn2df['filter_stubs'], d)
        image_trajectories(dtraj=dn2df['subtract_drift'], 
                           img_gfp=img_gfp, 
                           img_bright=img_bright, fig=None, ax=None).set(title='Drift substracted')
        plt.savefig(f"{cellcfg['plotp']}/image_trajectories_{dn2plotp_suffix['subtract_drift']}")

    dn2df['distance']=get_distance_from_centroid(dn2df['subtract_drift' if subtract_drift else 'filter_returns'],**params['get_distance_from_centroid'])
    dn2df['distance']=get_distance_travelled(t_cor=dn2df['distance'])
    
    for k in dn2df:
        to_table(dn2df[k].reset_index(drop=True),dn2dp[k])
        
def apply_cellcfgp2distances(
    cellcfgp: str,
    ):
    """
    Wrapper around cellcfg2distances for multiprocessing.

    Args:
        cellcfgp (str): path to the configuration file.
    """
    celli='/'.join(dirname(cellcfgp).split('/')[-3:])
    print(f"processing {celli}: ");logging.info(celli)
    cellcfg=read_dict(cellcfgp)
    cellcfg2distances(cellcfg,
                      test=cellcfg['test'],
                      force=cellcfg['force'])
    

def get_distance_from_centroid(
    df1: pd.DataFrame,
    center: list=[75,75],
    ) -> pd.DataFrame:
    """
    Get distance from the centroid.

    Args:
        df1 (pd.DataFrame): input dataframe.
        center (list, optional): center point. Defaults to [75,75].

    Returns:
        pd.DataFrame: output dataframe.
    """
    # get distance from center
#     from scipy.spatial import distance
    if 'distance effective from centroid' in df1:
        df1=df1.drop(df1.filter(like='distance effective from centroid',axis=1).columns.tolist(),axis=1)
    df1['distance effective from centroid per frame']=df1.apply(lambda x: distance.euclidean(center,[x['x'],x['y']]),axis=1)
    df=df1.groupby('particle',as_index=False).agg({'distance effective from centroid per frame':[np.min,np.max]})
    df=flatten_columns(df)
    df=df.rename(columns={c:c if not 'per frame a' in c else c.replace('per frame a','') for c in df})
#     print(df.columns)
    df['distance effective from centroid']=df['distance effective from centroid max']-df['distance effective from centroid min']
    return df1.merge(df,on='particle',how='left')

## distances
def distance_effective(
    particle,
    frame1,
    frame2,
    t_cor: pd.DataFrame,
    ) -> float:
    """
    Effective distance between frames.

    Args:
        particle : particle
        frame1 (np.array): a frame.
        frame2 (np.array): another frame.
        t_cor (pd.DataFrame): t_cor.

    Returns:
        float: distance
    """
    from scipy.spatial import distance
    a=t_cor.loc[((t_cor['particle']==particle) & (t_cor['frame']==frame1)),['x','y']]
    b=t_cor.loc[((t_cor['particle']==particle) & (t_cor['frame']==frame2)),['x','y']]
    return distance.euclidean(a.values, b.values)

def get_distance_travelled(
    t_cor: pd.DataFrame,
    ) -> pd.DataFrame:
    """
    Distance travelled.

    Args:
        t_cor (pd.DataFrame): input dataframe.

    Returns:
        pd.DataFrame: output dataframe.
    """
    for f1,f2 in zip(list(range(0,t_cor['frame'].max())),
                list(range(1,t_cor['frame'].max()+1))):
        for p in t_cor['particle'].unique():
            a=t_cor.loc[((t_cor['particle']==p) & (t_cor['frame']==f1)),['x','y']]
            b=t_cor.loc[((t_cor['particle']==p) & (t_cor['frame']==f2)),['x','y']]
            if len(a)!=0 and len(b)!=0:
                t_cor.loc[((t_cor['particle']==p) & (t_cor['frame']==f2)),'distance']=distance.euclidean(a.values, b.values)
    t_cor_distance=t_cor.groupby('particle').agg({'distance':sum})

    t_cor_rangeframes=t_cor.groupby('particle').agg({'frame':[min,max]})
    t_cor_rangeframes=flatten_columns(t_cor_rangeframes)
    t_cor_rangeframes['distance effective']=t_cor_rangeframes.apply(lambda x : distance_effective(x.name,x['frame min'],x['frame max'],t_cor) ,axis=1)

    t_cor=t_cor.merge(t_cor_distance,
                left_on='particle',right_index=True,suffixes=[' delta',' total'])
    t_cor=t_cor.merge(t_cor_rangeframes,
                left_on='particle',right_index=True)
    t_cor['distance delta']=t_cor['distance delta'].fillna(0)
    def get_distance_total_per_frame(df):
        df['distance total per frame']=df.apply(lambda x : df.set_index('frame').loc[[i for i in range(df['frame'].min(),int(x['frame'])+1)],'distance delta'].sum() ,axis=1)
        return df

    t_cor=t_cor.groupby('particle').apply(get_distance_total_per_frame)
    t_cor['distance effective per frame']=t_cor.apply(lambda x : distance_effective(particle=x['particle'],frame1=x['frame min'],frame2=x['frame'],t_cor=t_cor) ,axis=1)
    t_cor['intensity']=t_cor['mass']        
    return t_cor


#!/usr/bin/env python
"""Visualization of the images."""

import numpy as np
import matplotlib.pyplot as plt

def image_background(
    img_region=None,
    img=None,
    cmap: str='binary_r',
    alpha=1,
    linewidths=1,
    colors='cyan',
    kws_region={},
    show_scalebar=None,
    scalebar_color: str='w',
    show_cbar: bool=True,
    test=False,
    ax=None,
    **kws_img,
    ) -> plt.Axes:
    """Plot the image, to be used as a background to the annotations.

    Args:
        img_region (_type_, optional): segmentation image. Defaults to None.
        img (_type_, optional): image with intensity values. Defaults to None.
        cmap (str, optional): colormap name. Defaults to 'binary_r'.
        alpha (int, optional): transparency. Defaults to 1.
        linewidths (int, optional): segmentation contour line width. Defaults to 1.
        colors (str, optional): color of the segmentation line. Defaults to 'cyan'.
        kws_region (dict, optional): parameters provided to the segmentation plot. Defaults to {}.
        show_scalebar (_type_, optional): show scale bar. Defaults to None.
        scalebar_color (str, optional): color of the scale bar. Defaults to 'w'.
        show_cbar (bool, optional): show colorbar. Defaults to True.
        test (bool, optional): test-mode. Defaults to False.
        ax (_type_, optional): subplot object. Defaults to None.

    Keyword Args:
        parameters provided to the `plt.imshow`.
    Returns:
        plt.Axes
    """
    if cmap=='gfp':
        from roux.viz.colors import make_cmap
        cmap=make_cmap(
            ["#000000",'#83f52c'],
            N=50,
            )        
    ax=plt.subplot(111) if ax is None else ax
    if not img is None:
        ax_img=ax.imshow(
            img,
            cmap=cmap,
            alpha=alpha,
            **kws_img,
            )
    if not img_region is None:
        ax.contour(
            img_region, [0.5], 
            linewidths=linewidths,
            linestyles='dashed',
            colors=colors,
            **kws_region,
            )
    if not show_scalebar is None:
        ax=annotate_scalebar(
            ax,
            img=img,
            pixels=list(show_scalebar.values())[0],
            label=list(show_scalebar.keys())[0],
            )
    if not img is None and show_cbar:
        cax = ax.figure.add_axes([ax.get_position().x1+0.01,ax.get_position().y0,0.02,ax.get_position().height if ax.figure.get_size_inches()[1]<5 else (2/ax.figure.get_size_inches()[1])])
        cbar = ax.figure.colorbar(ax_img,cax=cax)
        cbar.ax.set_ylabel('Intensity')
        # cbar.ax.hist(img.ravel()) # show histogram in the colorbar
    ax.grid(test)
    ax.set_aspect('equal')
    return ax

def annot_cells(
    label_image,
    show_boxes: bool=False,
    ax: plt.Axes=None,
    )-> plt.Axes:
    """Annotate the cells on an image.

    Args:
        label_image (_type_): image with the labeled regions 
        show_boxes (bool, optional): show boxes around regions. Defaults to False.
        ax (plt.Axes, optional): plt.Axes. Defaults to None.

    Returns:
        plt.Axes
    """
    from skimage.measure import regionprops
    import matplotlib.patches as mpatches
    for region in regionprops(label_image):
        minr, minc, maxr, maxc = region.bbox
        ax.text(x=np.mean([minc,maxc]),y=np.mean([minr,maxr]),s=f"{region.label:.0f}",ha='center',va='center',color='r',zorder=10,size=10)
        if show_boxes:
            # draw rectangle around segmented coins
            rect = mpatches.Rectangle((minc, minr), maxc - minc, maxr - minr,
                                      fill=False, edgecolor='red', linewidth=2)
            ax.add_patch(rect)
    return ax

def image_regions_annotated(
    img_region,
    img,
    show_boxes: bool=False,
    **kws_img,
    ) -> plt.Axes:
    """
    Image with the annotated regions.
    Usage: for QC of the segmentation.

    Args:
        img_region (_type_): image with segmentated regions.
        img (_type_): image with intensity.
        show_boxes (bool, optional): whether to show the boxes around the regions. Defaults to False.

    Keyword Args:
        parameters provided to the `image_background` function.

    Returns:
        plt.Axes
    """
    ax=image_background(
        img_region=img_region,
        img=img,
        **kws_img,
        )
    annot_cells(
        label_image=img_region,
        show_boxes=show_boxes,
        ax=ax,
        )
    return ax


#!/usr/bin/env python

from os.path import splitext, join, exists, isdir,basename,abspath,dirname
from os import makedirs
import trackpy as tp
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import pims
# import pims_nd2
# import nd2reader
from glob import glob
import logging
logging.basicConfig(format='[%(asctime)s] %(levelname)s\tfrom %(filename)s in %(funcName)s(..): %(message)s',level=logging.DEBUG) # 

from os.path import dirname, basename
import matplotlib.pyplot as plt
import matplotlib.cm as cm

def plot_msd(
    imsd: pd.DataFrame,
    emsd: pd.DataFrame,
    scale: str="log",
    plot_fh: str=None,
    time_points_max: int = None,
    ax: plt.Axes=None,
    ) -> plt.Axes:
    """Plot MSD.

    Args:
       imsd (pd.DataFrame): mean squared displacement of each particle.
       emsd (pd.DataFrame): ensemble mean squared displacement of particles.
       scale (str, optional): axis scale. Defaults to "log".
       plot_fh (str, optional): output path. Defaults to None.
       params_msd (dict, optional): parameters of MSD. Defaults to { "mpp":0.0645, "fps":0.2, "max_lagtime":100 }.
       ax (plt.Axes, optional): subplot. Defaults to None.

    Returns:
       plt.Axes: subplot
       
    Examples:
        Calculate `time_points_max`:
        ```
            params_msd: dict={
                "mpp":0.0645,
                "fps":0.2,
                "max_lagtime":100
                }
            time_points_max=params_msd["fps"]*params_msd["max_lagtime"]
        ```
    """
    if ax is None:
        ax=plt.gca()
    if not time_points_max is None:
        imsd=imsd.head(time_points_max)
        emsd=emsd.head(time_points_max)

    imsd.plot(
        # x=imsd.index,
        legend=None,alpha=0.75,ax=ax)
    ax.set(ylabel=r'$\langle \Delta r^2 \rangle$ [$\mu$m$^2$]',
           xlabel='lag time $t$')
    if scale=="log":
        ax.set_xscale('log')
        ax.set_yscale('log')

    emsd.plot(
        # x=emsd.index,
        style='o',legend=None,ax=ax)
    if scale=="log":
        ax.set_xscale('log')
        ax.set_yscale('log')
    ax.set(ylabel=r'$\langle \Delta r^2 \rangle$ [$\mu$m$^2$]',
           xlabel='lag time $t$')
    plt.tight_layout()
    if not plot_fh is None: 
        ax.figure.savefig(plot_fh);
    return ax

def plot_emsd(
    expt_data: pd.DataFrame,
    color: str='k',
    scale: str="log",
    plot_fh: str=None,
    ax: plt.Axes=None,
    ) -> plt.Axes:
    """
    Plot ensemble mean squared displacement of particles.

    Args:
       expt_data (pd.DataFrame): input data.
       color (str, optional): color. Defaults to 'k'.
       scale (str, optional): scale of the axes. Defaults to "log".
       plot_fh (str, optional): output path. Defaults to None.
       ax (plt.Axes, optional): subplot. Defaults to None.

    Returns:
       plt.Axes: subplot
    """
    if ax is None:
        plt.figure(figsize=(6, 3))
        ax=plt.subplot(121)
    expt_data.plot(x=expt_data.index,style='o',c=color,ax=ax,markeredgecolor = 'none')
    if scale=="log":
        ax.set_xscale('log')
        ax.set_yscale('log')
    ax.set(ylabel=r'$\langle \Delta r^2 \rangle$ [$\mu$m$^2$]',
                xlabel='lag time $t$')
    ax.legend(loc = 'center left', bbox_to_anchor = (1.0, 0.5)) #bbox_to_anchor=(2.2, 1.0)
    plt.tight_layout()
    if not plot_fh is None: 
        ax.figure.savefig(plot_fh);
#         plt.clf();plt.close()
    return ax 

## kinetics 
def plot_kin(
    traj: pd.DataFrame,
    dparams: pd.DataFrame=None,
    ctime: str='time (s)',
    fit_eqn: str=None,
    ylabel: str='',
    label: str='',
    color: str='b',
    ax1: plt.Axes=None,
    plot_fh: str=None,
    ) -> plt.Axes:
    """Plot kinetics.

    Args:
        traj (pd.DataFrame): table containing the trajectories.
        dparams (pd.DataFrame, optional): table containing the parameters. Defaults to None.
        ctime (str, optional): time point. Defaults to 'time (s)'.
        fit_eqn (str, optional): fit equation name. Defaults to None.

    Returns:
        plt.Axes
    """
    traj_mean=traj.set_index(ctime)
    traj_mean=traj.T.mean()
    traj_std=traj.T.std()
    if fit_eqn=="logistic4":
        params=["Minimum asymptote","Hill's slope","Inflection point","Maximum asymptote"]
    elif fit_eqn=="logistic5":
        params=["Minimum asymptote","Hill's slope","Inflection point","Maximum asymptote","Asymmetry factor"]
    elif fit_eqn=='power':
        params=["power law exponent","amplitude"]
        
    if ax1 is None:
        plt.figure(figsize=(8,3))
        ax1=plt.subplot(131)
    ax1.set_xlabel(ctime)
    ax1.set_ylabel(ylabel)
    ax1.plot(traj_mean.index, traj_mean,'o',zorder=2, 
             markerfacecolor='none',markeredgecolor=color,alpha=0.5,
#             label='sd',
            )
    ax1.legend(label,loc='center', 
               bbox_to_anchor=(1.63, 0.5))
    ax1.fill_between(traj_std.index, \
                     traj_mean-traj_std, \
                     traj_mean+traj_std, \
                     color=color, alpha=0.2,zorder=0)
    
    if not dparams is None:
        params_mean=dparams.mean()
        params_std=dparams.std()
        p0=params_mean
    else:
        p0=np.zeros(len(params))
    if fit_eqn=="logistic4":
        plsq = leastsq(logistic4_residuals, p0, args=(traj_mean, traj_mean.index))
        ax1.plot(traj_mean.index.tolist(),logistic4_peval(traj_mean.index,plsq[0]),label=label,color=color,zorder=2)        
    elif fit_eqn=="logistic5":
        plsq = leastsq(logistic5_residuals, p0, args=(traj_mean, traj_mean.index))
        ax1.plot(traj_mean.index.tolist(),logistic5_peval(traj_mean.index,plsq[0]),label=label,color=color,zorder=2)        
    elif fit_eqn=="power":
        plsq = leastsq(power_residuals, p0, args=(traj_mean, traj_mean.index))
        ax1.plot(traj_mean.index.tolist(),power_peval(traj_mean.index,plsq[0]),label=label,color=color,zorder=2,lw=2)        
                
    ax1.set_ylim([0,ax1.get_ylim()[1]])
    plt.tight_layout()
    if plot_fh!=None:
        plt.savefig(plot_fh, format='pdf')
    return ax1

def plot_kin_all(
    expt_dh: str,
    imsd_fhs: list,
    ):
    """
    Plot multiple kinetics plots.
    """
    fig=plt.figure(figsize=(8,3))
    colors=cm.rainbow(np.linspace(0,1,len(imsd_fhs)))
    ylimmx=[]
    for i,imsd_fh in enumerate(imsd_fhs):
        imsd_flt_fh='%s.imsd_flt' % imsd_fh
        imsd_flt=pd.read_csv(imsd_flt_fh).set_index('lag time [s]')
        params_flt_fh='%s.params_flt' % imsd_fh
        params_flt=pd.read_csv(params_flt_fh)
        ax1=plt.subplot(131)
        ax1=plot_kin(imsd_flt,
                 params_flt.loc[:,['power law exponent','power law constant']],
                 ctime='lag $t$',
                 fit_eqn='power',
                 smp_ctrl=None,
                 ylabel='MSD ($\mu m^{2}$)',
                 color=colors[i],
                 ymargin=0.2,
                 ax1=ax1,
                 label=basename(imsd_flt_fh).split('_replicate', 1)[0].replace('_',' '),
                 plot_fh=None)
        ylimmx.append(np.max(imsd_flt.max()))
    ax1.set_ylim([0,np.max(ylimmx)])
    ax1.legend(bbox_to_anchor=[1,1,0,0],loc=2)
    fig.savefig('%s/plot_flt_%s.pdf' % (expt_dh,dirname(expt_dh)))

    
def plot_trajectories_stats(
    df: pd.DataFrame,
    coly: str,
    colx: str='frame',
    rescalex: bool=True,
    label: str=None,
    axvlinex=None,
    params_plot: dict={'color':'k','alpha':0.5},
    fig=None,
    ax: plt.Axes=None,
    ) -> plt.Axes:
    """
    Plot statistics of the particle trajectories.
    """
    fig=plt.figure() if fig is None else fig
    axin=not ax is None
    ax=plt.subplot() if ax is None else ax
    label=label if not hasattr(df,'label') else df.name       
    df=df.sort_values(colx)
    if rescalex:
        xmin=df[colx].min()
        df[colx]=range(len(df)) 
    ax=df.plot(x=colx,y=coly,ax=ax,label=label,**params_plot)
    ax.set_xlim(df[colx].min(),df[colx].max()+2)
    ax.set_ylim(df[coly].min(),df[coly].max())    
    if not axvlinex is None:
        ax.axvline(axvlinex-(xmin if rescalex else 0),
               label='inflection point',color='gray')
    ax.set_xlabel(colx);ax.set_ylabel(coly)
    return ax    

## on a cell image
def plot_trajectories(
    traj,
    image,
    colorby: str='particle',
    mpp: float=None,
    label: str=False,
    cmap: str=None,
    t_column: str=None,
    pos_columns: list=None,
    plot_style: dict={},
    params_text: dict={'ha':'center','va':'center'},
    ax: plt.Axes=None,
    **kwargs,
    ) -> plt.Axes:
    """Plot traces of trajectories for each particle.
    Optionally image it on a frame from the video.

    Parameters
    ----------
    traj : DataFrame
        The DataFrame should include time and spatial coordinate columns.
    colorby : {'particle', 'frame'}, optional
    mpp : float, optional
        Microns per pixel. If omitted, the labels will have units of pixels.
    label : boolean, optional
        Set to True to write particle ID numbers next to trajectories.
    image : ndarray, optional
        Background image, default None
    cmap : colormap, optional
        This is only used in colorby='frame' mode. Default = mpl.cm.winter
    ax : matplotlib axes object, optional
        Defaults to current axes
    t_column : string, optional
        DataFrame column name for time coordinate. Default is 'frame'.
    pos_columns : list of strings, optional
        Dataframe column names for spatial coordinates. Default is ['x', 'y'].
    plot_style : dictionary
        Keyword arguments passed through to the `Axes.plot(...)` command

    Returns
    -------
    Axes object
    
    See Also
    --------
    plot_traj3d : the 3D equivalent of `plot_traj`
    """
    import matplotlib as mpl
    import matplotlib.pyplot as plt
    from matplotlib.collections import LineCollection
    from roux.lib.plot.colors import get_cmap_subset
#     if cmap is None:
#         cmap = get_cmap_subset('binary',vmin=0.15,vmax=0.05)
    if t_column is None:
        t_column = 'frame'
    if pos_columns is None:
        pos_columns = ['x', 'y']
    if len(traj) == 0:
        raise ValueError("DataFrame of trajectories is empty.")
    # Background image
    ax=plt.subplot() if ax is None else ax
    ax.imshow(image,cmap=get_cmap_subset('binary',vmin=0.2,vmax=0))
    # Trajectories
    ax.set_xlim(0,image.shape[0])
    ax.set_ylim(0,image.shape[1])
    ax.set_ylim(ax.get_ylim()[::-1])
    return ax


#!/usr/bin/env python
"""
## Development

### release new version

    git commit -am "version bump";git push origin master
    python setup.py --version
    git tag -a v$(python setup.py --version) -m "upgrade";git push --tags

"""

import sys
if (sys.version_info[0]) != (3):
    raise RuntimeError('Python 3 required ')

import setuptools

with open('README.md', 'r') as fh:
    long_description = fh.read()
    
requirements = {
'base':[
    'roux>=0.0.9', # helper functions
    'scikit-image', # image processing.
    'argh', # for command-line convinience
    'seaborn',
    ],
'spt':[
    'pims==0.4.1', 
    'trackpy', # for tracking the particles
    'numba', # required by trackpy for faster processing
],
## development and maintenance
'dev':[
    'pytest',
    'jupyter','ipywidgets','ipykernel',
    # 'sphinx','recommonmark',
    'black',
    'coveralls == 3.*',
    'flake8',
    'isort',
    'pytest-cov == 2.*',
    'testbook',
    'papermill',
    'lazydocs', 'regex', ## docs
],
}

extras_require={k:l for k,l in requirements.items() if not k=='base'}
## all: extra except dev
extras_require['all']=[l for k,l in extras_require.items() if not k=='dev']
### flatten
extras_require['all']=[s for l in extras_require['all'] for s in l]
### unique
extras_require['all']=list(set(extras_require['all']))

# main setup command
setuptools.setup(
    name='htsimaging',
    version='1.0.4',
    description='High-Throughput Single-cell Imaging analysis.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='http://github.com/rraadd88/htsimaging',
    author='rraadd88',
    author_email='rohanadandage@gmail.com',
    license='General Public License v. 3',
    packages=setuptools.find_packages('.',exclude=['test', 'unit','deps', 'data']),
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=requirements['base'],
    extras_require=extras_require,
    # entry_points={
    # 'console_scripts': ['htsimaging = htsimaging.run:parser.dispatch',],
    # },
    python_requires='>=3.7, <4',
)


#!/usr/bin/env python
"""Utilities for the image processing."""

import logging

import numpy as np
import pandas as pd

from skimage.segmentation import random_walker
from skimage import restoration,filters,measure

def filterframe(
    frame,
    cutoff: float=0,
    ):
    """
    Filter a frame.

    Args:
        frame (array-like): input frame.
        cutoff (float, optional): cutoff on the values. Defaults to 0.

    Returns:
        array-like: filtered frame.
    """
    frame_cleand=np.array(frame)
    frame_cleand[frame_cleandi<cutoff]=0
    return frame_cleaned

def filterframes(
    frames: list,
    cutoff: float=0,
    ) -> list:
    """Filter the frames.

    Args:
        frames (list): list of frames.
        cutoff (float, optional): threshold on the values. Defaults to 0.

    Returns:
        list: list of frames.
    """
    frames_cleaned=[]
    for framei in frames:
        frames_cleaned.append(filterframe(framei,cutoff))
    return frames_cleaned

def get_data_by_regions(
    regions: list,
    img=None,
    prop_type: str='area',
    ) -> pd.DataFrame:
    """Get properties by regions.

    Args:
        regions (list): list of the regions.
        img (array-like, optional): input image. Defaults to None.
        prop_type (str, optional): property type. Defaults to 'area'.

    Returns:
        pd.DataFrame: output dataframe.
    """
    if prop_type=='mean_intensity':
        ValueError("arg img is required")
    regions_props= measure.regionprops(regions.astype(int),intensity_image=img)
    regions_lbls = np.array([prop.label for prop in regions_props])
    if prop_type=='area':
        regions_props_selected = np.array([getattr(prop,prop_type)  for prop in regions_props])
    df1=pd.DataFrame({'labels':regions_lbls,
                      prop_type:regions_props_selected,
                     })
    return df1

def filter_regions(
    regions: list,
    kind='labeled',
    img=None,
    prop_type: str='area',
    mn: float=0,
    mx: float=0,
    test: bool=False,
    plotp: str=None,
    ) -> list:
    """
    Filter regions.

    Args:
        regions (np.array): segmented image, labeled with `measure.label(regions)`.
        img (array-like, optional): image. Defaults to None.
        prop_type (str, optional): property type. Defaults to 'area'.
        mn (float, optional): minimum value. Defaults to 0.
        mx (float, optional): maximum value. Defaults to 0.
        test (bool, optional): test the filtering. Defaults to False.
        plotp (str, optional): path to the plot. Defaults to None.

    Raises:
        ValueError: img is required if filtering is t be done by the intensity.

    Returns:
        list: list of the filtered regions.

    Notes:
        Prefer `regionprops_table` which is new in sklean's version 0.16.
    """
    assert mn<=mx
    if prop_type=='mean_intensity' and img is None:
        raise ValueError("img is required")
    if kind!='labeled' and len(set(regions.ravel()))==2:
        logging.warning('image contains 2 unique values, make sure it is labeled `measure.label(regions)` and not binary.')
    regions_props= measure.regionprops(regions.astype(int),intensity_image=img)
    regions_lbls = np.array([prop.label for prop in regions_props])
    if prop_type=='centroid_x':
        regions_props_selected = np.array([prop.centroid[0] for prop in regions_props])     
    elif prop_type=='centroid_y':
        regions_props_selected = np.array([prop.centroid[1] for prop in regions_props])     
    else:
        regions_props_selected = np.array([getattr(prop,prop_type)  for prop in regions_props])
        
    regions_filtered_lbls = regions_lbls[np.where((regions_props_selected<=mx) & (regions_props_selected>=mn))[0]]
    regions_filtered=np.zeros(regions.shape)
    for lbli in regions_filtered_lbls:
        regions_filtered[np.where(regions==lbli)]=lbli
    if test:
        logging.info(f"np.unique(regions_props_selected.ravel()) = {np.unique(regions_props_selected.ravel())}")
        logging.info(f"regions_lbls = {regions_lbls}")
        logging.info(f"regions_filtered_lbls = {regions_filtered_lbls}")
        import matplotlib.pyplot as plt
        fig,axs=plt.subplots(1,2)
        ax=axs[0]
        print(np.unique(regions_props_selected.ravel()))
        ax.hist(regions_props_selected)
        ax.set(title=prop_type)
        ax=axs[1]
        if not img is None:
            ax.imshow(img, cmap=plt.cm.gray, interpolation='nearest')        
        ax.contour(regions, [0.5], linewidths=1.2, colors='r')
        ax.contour(regions_filtered, [0.5], linewidths=1.2, colors='g')
        plt.tight_layout()
        if not plotp is None:
            makedirs(dirname(plotp),exist_ok=True)
            plt.savefig(plotp)
    return regions_filtered

def smoothen(
    img,
    ):
    """Smoothen the image.

    Args:
        img (array-like): input image.

    Returns:
        array-like: output image
    """
    denoised=restoration.denoise_bilateral(img.astype('uint16'), sigma_range=0.01, sigma_spatial=15)
    smoothened = filters.median(denoised,np.ones((4,4)))
    return smoothened

def smoothenframes(
    frames: list,
    ) -> list:
    """Smoothen the images.

    Args:
        frames (list): list of frames.

    Returns:
        list: list of frames.
    """
    frames_cleaned=[]
    for framei in frames:
        frames_cleaned.append(smoothen(np.array(framei)))
    return frames_cleaned
        
def get_regions(
    img,
    ):
    """Get regions.

    Args:
        img (array-like): input image
    """
    denoised=restoration.denoise_bilateral(img.astype('uint16'), 
#                                            sigma_range=0.01, 
                                           sigma_spatial=15,
                                          multichannel=False)
    smoothened = filters.median(denoised,np.ones((4,4)))
    markers = np.zeros(smoothened.shape, dtype=np.uint)
# otsu works only for only for multi-channel images     
#     markers[smoothened < filters.threshold_otsu(smoothened)] = 1
#     markers[smoothened > filters.threshold_otsu(smoothened)] = 2
    markers[smoothened < filters.median(smoothened)] = 1
    markers[smoothened > filters.median(smoothened)] = 2

    labels = random_walker(smoothened, markers, beta=10, mode='bf')
    regions= measure.label(labels)
    return regions, denoised, smoothened,markers

def raw2phasecorr(
    arr_list: list,
    clip: int=0,
    ) -> list: #cv
    """Correct for the relative translative offset by phase correlation between images.

    Args:
        arr_list (list): list of frames.
        clip (int, optional): threshold. Defaults to 0.

    Returns:
        list: output frames.
    """
    import cv2
    cx = 0.0
    cy = 0.0
    stb_arr_list=[]
    prev_frame = arr_list[0]
    prev_image = np.float32(restoration.denoise_tv_chambolle(prev_frame.astype('uint16'), weight=0.1, multichannel=True)) #ref
    for frame in arr_list:           
        image = np.float32(restoration.denoise_tv_chambolle(frame.astype('uint16'), weight=0.1, multichannel=True))
        # TODO: set window around phase correlation
        dp = cv2.phaseCorrelate(prev_image, image)
        cx = cx - dp[0]
        cy = cy - dp[1]
        xform = np.float32([[1, 0, cx], [0, 1, cy]])
        stable_image = cv2.warpAffine(frame.astype('float32'), xform, dsize=(image.shape[1], image.shape[0]))
        prev_image = image
        #clip sides
        ht,wd=np.shape(stable_image)
#         clip=0.125 #0.25
        lt=int(wd*clip)
        rt=int(wd-wd*clip)
        up=int(ht*clip)
        dw=int(ht-ht*clip)
        stable_image_clipped=stable_image[up:dw,lt:rt]
        stb_arr_list.append(stable_image_clipped)
    return stb_arr_list

def phasecorr_with(
    imlist: list,
    imlist2: list=None,
    clip: int=0,
    ): #cv  [rowini,rowend,colini,colend]
    """Correct for the relative translative offset by phase correlation with a set of given images. 

    Args:
        imlist (list): test images
        imlist2 (list, optional): reference images. Defaults to None.
        clip (int, optional): threshold. Defaults to 0.

    Returns:
        list: corrected images.
    """
    import cv2
    cx = 0.0
    cy = 0.0            
    imlist_stb=[]
    if imlist2!=None:
        imlist2_stb=[]

    imi=0
    im_prev = imlist[0]
    im_denoised_prev = np.float32(restoration.denoise_tv_chambolle(im_prev.astype('uint16'), weight=0.1, multichannel=True)) #ref
    for im in imlist:           
        im_denoised = np.float32(restoration.denoise_tv_chambolle(im.astype('uint16'), weight=0.1, multichannel=True))
        # TODO: set window around phase correlation
        dp = cv2.phaseCorrelate(im_denoised_prev, im_denoised)
        cx = cx - dp[0]
        cy = cy - dp[1]
        xform = np.float32([[1, 0, cx], [0, 1, cy]])
        im_stb = cv2.warpAffine(im.astype('float32'), xform, dsize=(im_denoised.shape[1], im_denoised.shape[0]))
        imlist_stb.append(imclipper(im_stb,clip))

        if imlist2!=None:
            im2=imlist2[imi]
            im2_stb=cv2.warpAffine(im2.astype('float32'), xform, dsize=(im_denoised.shape[1], im_denoised.shape[0]))
            imlist2_stb.append(imclipper(im2_stb,clip))

        im_denoised_prev = im_denoised
        imi+=1
    if imlist2!=None:
        return imlist_stb,imlist2_stb
    else:
        return imlist_stb
    
def imlistcropper(
    imlist: list,
    loci: int,
    ) -> list:
    """
    Crop a list of images.

    Args:
        imlist (list): list of images.
        loci (int): locations.

    Returns:
        list: output images.
    """
    rowini,rowend,colini,colend=loci
    imlist_region=[]
    for im in imlist:
        imlist_region.append(im[rowini:rowend,colini:colend])
    return imlist_region


def imclipper(
    im_stb,
    clip: float,
    ):
    """Crop an image.

    Args:
        im_stb (array-like): input image
        clip (float): threshold.
    """
    ht,wd=np.shape(im_stb)
#         clip=0.125 #0.25
    lt=int(wd*clip)
    rt=int(wd-wd*clip)
    up=int(ht*clip)
    dw=int(ht-ht*clip)
    im_stb_clipped=im_stb[up:dw,lt:rt]
    return im_stb_clipped

## features
def get_cellprops(
    regions,
    intensity_imgtype2img,
    properties=['area',
        'bbox_area',
        'convex_area',
        'eccentricity',
        'equivalent_diameter',
        'euler_number',
        'extent',
        'filled_area',
        'label',
        'major_axis_length',
        'max_intensity',
        'mean_intensity',
        'min_intensity',
        'minor_axis_length',
        'orientation',
        'perimeter',
        'solidity',
        'centroid',
                ],
    ) -> pd.DataFrame:
    """Get cell properties.

    Args:
        regions (list): regions in a frame.
        intensity_imgtype2img (_type_): intensities.
        properties (list, optional): _description_. Defaults to ['area', 'bbox_area', 'convex_area', 'eccentricity', 'equivalent_diameter', 'euler_number', 'extent', 'filled_area', 'label', 'major_axis_length', 'max_intensity', 'mean_intensity', 'min_intensity', 'minor_axis_length', 'orientation', 'perimeter', 'solidity', 'centroid', ].

    Returns:
        pd.DataFrame: output dataframe.
    """
    dn2df={}
    for imgtype in intensity_imgtype2img:
        df=pd.DataFrame(measure.regionprops_table(regions.astype(int),
                                           intensity_image=intensity_imgtype2img[imgtype],
                                           properties=properties))
        df.index=df.index+1
        df.index.name='cell #'
        dn2df[imgtype]=df.melt(ignore_index=False,var_name='property').reset_index()
    df=pd.concat(dn2df,axis=0,names=['image type']).reset_index(0)
    return df

def get_signal_summary_by_roi(
    cellframes: list,
    xy_center: tuple=None,
    width: int=20,
    fun_summary_frame: str='min',
    fun_summary_frames: str='median',
    ):
    """Place of the roi in the image is defined by

    Args:
        cellframes (list): list of frames.
        xy_center (tuple, optional): position of the center. Defaults to None.
        width (int, optional): width of the ROI. Defaults to 20.
        fun_summary_frame (str, optional): function to summarise a frame. Defaults to 'min'.
        fun_summary_frames (str, optional): function to summarise a list of frames. Defaults to 'median'.

    Returns:
        summary value
    """
    width_half=int(width*0.5)
    frame=cellframes[0]
    if xy_center is None:
        xy_center=[int(i/2) for i in cellframes[0].shape]
    signal_cytoplasm=getattr(np,fun_summary_frames)([getattr(np,fun_summary_frame)(frame[xy_center[1]-width_half:xy_center[1]+width_half,xy_center[0]-width_half:xy_center[0]+width_half]) for frame in cellframes])
    return int(signal_cytoplasm)


#!/usr/bin/env python
"""I/O"""

import os
from os import makedirs
from os.path import exists,splitext,basename

import numpy as np
import matplotlib.pyplot as plt

import subprocess
from skimage import io

## image io
def read_image(
    imp: str
    ):
    """Read image.

    Args:
        imp (str): path to the image file.

    Returns:
        np.array
        
    TODOs: 
        For a tiff file:
        from skimage.external import tifffile
    """
    from skimage import io
    if isinstance(imp,(str)) and not isinstance(imp,(list)):
        if not imp.endswith('.npy'):
            im=io.imread(imp,as_gray=True)
        else:
            im=np.load(imp)
    else:
        im=imp
    return im

## make a video
def arr2vid(
    arr_list: list,
    regions: list,
    kins_mean: float,
    vid_fh: str,
    xpixels: list,
    ypixels: list,
    dpi: int = 100,
    ) -> str:
    """From array to video.

    Args:
        arr_list (list): list of frames.
        regions (list): regions 
        kins_mean (float): kinetics 
        vid_fh (str): video file path
        xpixels (list): pixels allong x-axis.
        ypixels (list): pixels allong y-axis.
        dpi (int, optional): DPI resolution. Defaults to 100.

    Returns:
        str: path of the video
    """
    png_dh=os.path.splitext(vid_fh)[0]
    if not os.path.exists(png_dh):
        try:
            os.makedirs(png_dh)
        except :
            logging.warning("error data_lbl")
    fig = plt.figure(figsize=(ypixels/dpi*2, xpixels/dpi), dpi=dpi)
    ax_img = plt.subplot(1,2,1)
    ax_kin = plt.subplot(1,2,2)
    ax_img.set_axis_off()
    ax_img.set_aspect('equal')
    for i in range(len(arr_list)):
        ax_img.imshow(arr_list[i],cmap='gray',animated=True)
        ax_img.contour(regions, [0.25], linewidths=1.2, colors='r',animated=False)
        if len(kins_mean.columns)>1:
            kins_mean.plot(x='time',legend=False,ax=ax_kin)
            ax_kin.plot(kins_mean['time'],kins_mean.drop(['time'], axis=1).mean(axis=1),lw=6,color='k')
            ax_kin.set_ylabel("Fluorescence Intensity (FU)")
        ax_kin.set_xlim([kins_mean.loc[0,'time'],kins_mean.loc[len(kins_mean)-1,'time']])
        ax_kin.axvline(kins_mean.loc[i,'time'], color='r', linestyle='--',lw=2)
        fig.subplots_adjust(wspace=.4)
        fig.savefig(png_dh+'/%02d.png' % i)
        ax_kin.clear()
    bash_command=("ffmpeg -f image2 -r 4 -i "+png_dh+"/%02d.png -vcodec mpeg4 -y "+vid_fh)
    subprocess.Popen(bash_command, shell=True, executable='/bin/bash')
    return vid_fh
    
def makevid(
    gfp_list_stb: list,
    brf_list_stb: list,
    cmap_gfp: str,
    cmap_brf: str,
    vid_fh: str,
    conditionn: int=None,
    interval=None,
    dpi : int= 300,
    ) -> str:
    """Convert to a video.

    Args:
        gfp_list_stb (list): channel (e.g. GFP) images.  
        brf_list_stb (list): segmented (e.g. bright-field) images. 
        cmap_gfp (str): colormap for the channel images.
        cmap_brf (str): colormap for the segmented images.
        vid_fh (str): path to the video file.
        conditionn (int, optional): title. Defaults to None.
        interval (_type_, optional): inerval of the frames. Defaults to None.
        dpi (int, optional): DPI resolution. Defaults to 300.

    Returns:
        str: path to the video file.
    """
    png_dh=splitext(vid_fh)[0]
    if not exists(png_dh):
        makedirs(png_dh)
    for i in range(len(gfp_list_stb)):
        fig = plt.figure(dpi=dpi)
        ax=plt.subplot(111)    
        ax.imshow(gfp_list_stb[i],cmap=cmap_gfp,zorder=1,alpha=0.5)
        ax.imshow(brf_list_stb[i],cmap=cmap_brf,zorder=0,alpha=0.5)
        ax.text(2,2,conditionn,color='r',fontsize=20,)
        ax.text(2,np.shape(gfp_list_stb[i])[1]*0.95,"time: %03d s" % ((i+1)*interval),color='b',fontsize=20,)
#         fig.subplots_adjust(wspace=.4)
        plt.tight_layout()
        plt.axis('off')
        fig.savefig('%s/%03d.png' % (png_dh,i))
        plt.close(fig)
    bash_command=("ffmpeg -f image2 -r 20 -i "+png_dh+"/%03d.png -vcodec mpeg4 -y "+vid_fh)
    print(vid_fh)
    subprocess.Popen(bash_command, shell=True)
    return vid_fh

def nd2arr_list(
    nd_dh: str=None,
    nd_fns: list=[],
    nd_fh: str=None,
    ) -> list:
    """Raw image to list of arrays.

    Args:
        nd_dh (str, optional): directory containing raw files e.g. nd2. Defaults to None.
        nd_fns (list, optional): file names. Defaults to [].
        nd_fh (str, optional): path to the files. Defaults to None.

    Returns:
        list: list of arrays
    """
    import nd2reader
    arr_list=[]
    if nd_fh is None:
        for nd_fn in nd_fns:
            nd = nd2reader.Nd2("%s/%s" % (nd_dh,nd_fn))
            for ndi in nd:
                arr_list.append(np.array(ndi))
            del nd
        return arr_list
    elif not nd_fh is None:
        nd = nd2reader.Nd2("%s" % (nd_fh))
        for ndi in nd:
            arr_list.append(np.array(ndi))
        return arr_list

def to_csv(
    fh_xls='../test/test.xlsx',
    nd2_dh="/media/Transcend/20160219_000356_267",
    cores=16,
    ):
    """
    Convert nd2 files to csv using parallel processing.

    Args:
        fh_xls (str, optional): metadata file. Defaults to '../test/test.xlsx'.
        nd2_dh (str, optional): path of the directory containing raw images. Defaults to "/media/Transcend/20160219_000356_267".
        cores (int, optional): number of cores. Defaults to 16.
    """
    import nd2reader
    import pandas as pd
    data_job=pd.read_excel(fh_xls,'JobView')
    nd_fhs=[ nd2_dh+'/'+str(fn) for fn in data_job['File Name'].unique()]

    # convert t0 csvs
    def nd2csv(
        nd_fh: str,
        ):
        nd = nd2reader.Nd2(nd_fh)
        frame_cnt=0
        for frame in nd:
            np.savetxt("%s/%s%02d.csv" % (nd2_dh,basename(nd_fh),frame_cnt), frame, delimiter=",")
            frame_cnt+=1
        nd.close
    from multiprocessing import Pool
    pool_data_lbl2data_fit=Pool(processes=int(cores)) 
    pool_data_lbl2data_fit.map(nd2csv,nd_fhs)
    pool_data_lbl2data_fit.close(); pool_data_lbl2data_fit.join()